# -*- coding: utf-8 -*-

try:
    import argparse
    import ConfigParser
    import contextlib
    import itertools
    import json
    import logging
    import logging.handlers
    import os
    import shutil
    import subprocess
    import sys
    import time

    from collections import OrderedDict 

    from salt.exceptions import CommandExecutionError
    IMPORTED = True
except ImportError:
    IMPORTED = False


def __virtual__():
    if not IMPORTED:
        return (
            False,
            'Failed to import required modules for {}'.format(__name__)
        )
    return True


def disable(force=False):
    try:
        _setup_logger(LOG_FILE)
    except Exception as error:
        return _salt_fail(error)

    try:
        return _disable_monitoring(force)
    except Success as complete:
        _salt_succeed(str(complete))
    except Exception as error:
        return _salt_fail(error)


def check():
    try:
        _setup_logger(LOG_FILE)
    except Exception as error:
        return _salt_fail(error)

    try:
        _check_disable_is_necessary()
    except Success as complete:
        _salt_succeed(str(complete))
    except Exception as error:
        return _salt_fail(error)

    return _salt_succeed('Monitoring is correctly disabled')


#####################
# Business logic
#####################

PROCESS_LIST_JSON = '/etc/128technology/processManagerProcessList.json'
INFLUX_CONFIG = '/etc/128technology/influxdb/influxdb.conf'

INFLUXD = '/usr/bin/influxd'
INFLUX_CLI = '/usr/bin/influx'
INFLUXDB_DIR = '/var/lib/128technology/influxdb'

MONITORING_ENV_VAR = 'INFLUXDB_MONITOR_STORE_ENABLED'
MONITORING_ENV_VALUE = MONITORING_ENV_VAR + '=false'

LOG_FILE = '/var/log/128technology/influx_disable_monitoring.log'


class FatalError(Exception):
    pass


class SuccessfulStateReached(Exception):
    pass


class StepComplete(Exception):
    pass


class ForceRemovalRequired(Exception):
    pass


class Success(Exception):
    pass


class FailedToCreateLoggingDirectory(FatalError):
    def __init__(self):
        super(FailedToCreateLoggingDirectory, self).__init__(
            'The logging directory {} could not be created'.format(os.path.dirname(LOG_FILE))
        )


class ProcessListDoesNotExist(FatalError):
    def __init__(self):
        super(ProcessListDoesNotExist, self).__init__('The process list does not exist.') 

        
class ProcessListHasInvalidJson(FatalError):
    def __init__(self):
        super(ProcessListHasInvalidJson, self).__init__('The process list is not valid json') 


class ProcessListMalformed(FatalError):
    def __init__(self, entries):
        super(ProcessListMalformed, self).__init__(
            'The process list does not have the expected entries: {}'.format(entries)
        )

        
class ProcessListInvalidInfluxEntriesCount(FatalError):
    def __init__(self, count):
        super(ProcessListInvalidInfluxEntriesCount, self).__init__(
            'The process list has {} InfluxDB entries when 1 was expected'.format(count)
        ) 


class ProcessListRewriteFailed(FatalError):
    def __init__(self):
        super(ProcessListRewriteFailed, self).__init__('Failed while rewriting the process list') 


class InfluxdDoesNotExist(FatalError):
    def __init__(self):
        super(InfluxdDoesNotExist, self).__init__('{} does not exist. Is 128T installed?'.format(INFLUXD))


class InfluxShellNotInstalled(FatalError):
    def __init__(self):
        super(InfluxShellNotInstalled, self).__init__('{} does not exist. Is 128T installed?'.format(INFLUX_CLI))


class InfluxRunningWithNoConfig(FatalError):
    def __init__(self):
        super(InfluxRunningWithNoConfig, self).__init__(
            'Influx is running, but no config exists. This should be an unreachable state.'
        )


class InfluxDataExistsWithNoConfig(FatalError):
    def __init__(self):
        super(InfluxDataExistsWithNoConfig, self).__init__(
            'Influx has produced data, but no config exists. This should be an unreachable state.'
        )


class InfluxConfigHasInvalidIni(FatalError):
    def __init__(self):
        super(InfluxConfigHasInvalidIni, self).__init__(
            'The config in {} cannot be parsed'.format(INFLUX_CONFIG)
        )


class InfluxConfigRewriteFailed(FatalError):
    def __init__(self):
        super(InfluxConfigRewriteFailed, self).__init__(
            'Failed while rewriting the Influx config'
        )


class InfluxInstanceCouldNotBeLaunched(ForceRemovalRequired):
    def __init__(self):
        super(InfluxInstanceCouldNotBeLaunched, self).__init__(
            'Influx could not be launched to drop the monitoring database.'
        )


class InfluxRefusingConnections(ForceRemovalRequired):
    def __init__(self):
        super(InfluxRefusingConnections, self).__init__(
            'Influx is running but not accepting connections.'
        )


class DroppedInternalDatabaseStillExists(ForceRemovalRequired):
    def __init__(self):
        super(DroppedInternalDatabaseStillExists, self).__init__(
            'The monitoring database was dropped but exists again.'
        )


class NoSignOfInfluxHavingRun(SuccessfulStateReached):
    def __init__(self):
        super(NoSignOfInfluxHavingRun, self).__init__(
            'Influx is not running and has no config or data. No more changes necessary.'
        )

class InfluxInternalDatabaseDoesNotExist(SuccessfulStateReached):
    def __init__(self):
        super(InfluxInternalDatabaseDoesNotExist, self).__init__(
            'Influx is running with no monitoring database.'
        )


class ProcessListMonitoringAlreadyDisabled(StepComplete):
    def __init__(self):
        super(ProcessListMonitoringAlreadyDisabled, self).__init__(
            'Monitoring already disabled in the process list'
        )


class InfluxConfigMonitoringAlreadyDisabled(StepComplete):
    def __init__(self):
        super(InfluxConfigMonitoringAlreadyDisabled, self).__init__(
            'Monitoring already disabled in the Influx config'
        )


def _disable_monitoring(force=False):
    try:
        logging.info('Disabling Influx monitoring in the process list')
        _disable_influx_monitoring_in_process_list()
        logging.info('Influx monitoring disabled in the process list')
    except StepComplete as complete:
        logging.info(complete)
    except FatalError as error:
        _fail(error) 

    if _influx_is_running():
        with _suppress(Exception):
            if not _monitoring_disabled_in_active_influx():
                _succeed('The running Influx instance has no monitoring database')
    else:
        with _suppress(Exception):
            if not _internal_data_directory_exists():
                _succeed('No Influx instance is running and no monitoring data exists')

    try:
        logging.info('Disabling Influx monitoring in the Influx config')
        _disable_monitoring_in_influx_config()
        logging.info('Influx monitoring temporarily disabled in the Influx config')
    except StepComplete as complete:
        logging.info(complete)
    except SuccessfulStateReached as complete:
        _succeed(str(complete))
    except FatalError as error:
        _fail(error)


    force_remove = False

    try:
        logging.info('Dropping the monitoring database')
        _delete_internal_database()
        logging.info('Monitoring database dropped from Influx')
    except ForceRemovalRequired as error:
        if force:
            force_remove = True
        else:
            _fail(error)
    except (StepComplete, SuccessfulStateReached) as complete:
        _succeed(str(complete))
    except FatalError as error:
        _fail(error)


    if force_remove:
        try:
            logging.info(str(error) + ' Forcefully removing all data.')
            _delete_all_data()
            _succeed('All Influx data cleaned')
        except (StepComplete, SuccessfulStateReached) as complete:
            _succeed('All Influx data cleaned')
        except (FatalError, ForceRemovalRequired) as error:
            logging.info('Failed to remove all data')
            _fail(error)


def _check_disable_is_necessary():
    try:
        if not _monitoring_disabled_in_process_list():
            _fail_with_message('Monitoring is not disabled in the process list')
        logging.info('Monitoring is disabled in the process list')
    
        if not _monitoring_disabled_in_active_influx():
            _fail_with_message('There is a running Influx instance with monitoring enabled')

        logging.info('Monitoring is disabled for any active Influx instance')
    except Exception as error:
        _fail(error)

    logging.info('Monitoring is correctly disabled')


def _monitoring_disabled_in_process_list():
    process_list = _load_process_list()
    influx_process = _get_influx_process(process_list)
    try:
        return MONITORING_ENV_VALUE in influx_process['environment']
    except KeyError:
        return False



def _monitoring_disabled_in_active_influx():
    if not _influx_is_running():
        return not _internal_data_directory_exists()

    if not _can_connect_to_influx():
        raise InfluxRefusingConnections()
    
    return not _internal_database_exists()



def _internal_data_directory_exists():
    internal_data_exists = (
        os.path.exists(INFLUXDB_DIR + '/data/_internal')
        or os.path.exists(INFLUXDB_DIR + '/data/_internal')
    )

    if internal_data_exists:
        logging.debug('Influx monitoring data still exists')
    else:
        logging.debug('There is no monitoring data on disk')

    return internal_data_exists


def _setup_logger(log_file_path):
    _initialize_logger()
    _add_file_logger(log_file_path)
    _add_console_logging()


def _initialize_logger():
    logging.getLogger().setLevel(logging.DEBUG)


def _add_file_logger(log_file_path):
    log_dir = os.path.dirname(log_file_path)

    with _suppress(os.error):
        os.makedirs(log_dir)

    if not os.path.exists(log_dir):
        raise FailedToCreateLoggingDirectory()

    log_exists = os.path.exists(log_file_path)

    try:
        extended_formatter = logging.Formatter(
            '%(asctime)s - %(levelname)s - %(message)s'
        )

        file_handler = logging.handlers.RotatingFileHandler(
            log_file_path,
            maxBytes=2**20,
            backupCount=20
        )

        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(extended_formatter)

        logging.getLogger().addHandler(file_handler)

        if log_exists:
            file_handler.doRollover()

    except OSError as err:
        logging.critical('Failed to initialize the logger: {}'.format(err))
        _fail_with_message('Failed to initialize logging')


def _add_console_logging():
    console_handler = logging.StreamHandler(stream=sys.stdout)
    console_handler.setLevel(logging.INFO)

    simple_formatter = logging.Formatter('%(message)s')
    console_handler.setFormatter(simple_formatter)

    logging.getLogger().addHandler(console_handler)


def _fail(error):
    raise CommandExecutionError(str(error))


def _fail_with_message(message):
    raise CommandExecutionError(message)


def _salt_fail(error):
    logging.error(error)
    return { 'info': str(error) }


def _salt_succeed(message=None):
    if message is not None:
        logging.info(message)

    return {}


def _succeed(message=None):
    raise Success(message)


def _delete_internal_database():
    if _influx_is_running():
        logging.debug('Influx is running. Operating against the existing instance.')
        _drop_internal_database_from_running_influx()
        _restart_running_influx()
        _verify_internal_database_is_gone()
    else:
        logging.debug('Influx is not running. Launching a new instance to operate against.')
        with _influx_instance():
            _drop_internal_database_from_running_influx()
            _verify_internal_database_is_gone()


def _drop_internal_database_from_running_influx():
    if not _can_connect_to_influx():
        raise InfluxRefusingConnections()

    if _internal_database_exists():
        _drop_internal_database()


def _restart_running_influx():
    _kill_influx()

    if _can_connect_to_influx():
        _kill_conflux()
    else:
        raise InfluxRefusingConnections()


def _verify_internal_database_is_gone():
    if not _internal_database_exists():
        raise InfluxInternalDatabaseDoesNotExist()
    else:
        raise DroppedInternalDatabaseStillExists()



def _internal_database_exists():
    try:
        logging.debug('Retrieveing current databases')
        output = _run_influx_shell_command('show databases')
    except subprocess.CalledProcessError as error:
        logging.debug('Failed to retrieve current databases')
        return False
    else:
        result = '_internal' in output.split()
        if result:
            logging.debug('Influx contains the monitoring database')
        else:
            logging.debug('Influx does not contain the monitoring database')
        return result


def _delete_all_data():
    logging.info('Deleting all data (enabled by the "force" option)')
    if _influx_is_running():
        shutil.rmtree(INFLUXDB_DIR)
        _restart_running_influx()
    else:
        shutil.rmtree(INFLUXDB_DIR)


def _can_connect_to_influx():
    SECONDS_TO_WAIT_FOR_CONNECTION = 60

    for i in range(SECONDS_TO_WAIT_FOR_CONNECTION):
        with _suppress(subprocess.CalledProcessError):
            _run_influx_shell_command('show databases')
            logging.debug('Successfully connected to Influx')
            return True
        if i % 10 == 0:
            logging.debug(
                'No Influx connection on check ' +
                str(i) + '/' + str(SECONDS_TO_WAIT_FOR_CONNECTION)
            )
        time.sleep(1)
    return False


def _influx_is_running():
    SECONDS_TO_WAIT_FOR_INFLUX = 10
    with _devnull() as devnull:
        for i in range(SECONDS_TO_WAIT_FOR_INFLUX):
            with _suppress(subprocess.CalledProcessError):
                subprocess.check_call(['pgrep', 'influxd'], stdout=devnull, stderr=devnull)
                logging.debug('Influx is running.')
                return True

            logging.debug(
                'Influx not running on check ' +
                str(i + 1) + '/' + str(SECONDS_TO_WAIT_FOR_INFLUX)
            )
            time.sleep(1)

        logging.debug('Influx is not running.')
        return False


def _drop_internal_database():
    logging.debug('Dropping the monitoring database')
    with _suppress(subprocess.CalledProcessError):
        _run_influx_shell_command('drop database _internal')

    
def _kill_influx():
    _kill_process('influxd')


def _kill_conflux():
    _kill_process('conflux')


def _kill_process(process_name):
    with _suppress(subprocess.CalledProcessError):
        logging.debug('Killing process {}'.format(process_name))
        subprocess.check_call(['pkill', process_name])


def _run_influx_shell_command(cmd, database=None):
    if not os.path.exists(INFLUX_CLI):
        raise InfluxShellNotInstalled()

    command = [INFLUX_CLI, '-execute', cmd]

    if database is not None:
        command.extend(['-database', database])

    with _devnull() as devnull:
        return subprocess.check_output(command, stderr=devnull)


@contextlib.contextmanager
def _influx_instance():
    if not os.path.exists(INFLUXD):
        raise InfluxdDoesNotExist()

    influx_rc = None
    influx = [INFLUXD, '-config', INFLUX_CONFIG]

    with _devnull() as devnull:
        try:
            logging.debug('Launching an influx instance: ' + str(influx))
            process = subprocess.Popen(influx, stdout=devnull, stderr=devnull)

            # verify that influx does not exit immediately
            time.sleep(1)
            influx_rc = process.poll()
            if influx_rc is not None:
                raise InfluxInstanceCouldNotBeLaunched()

            yield process
        finally:
            logging.debug('Stopping the influx instance')
            if influx_rc is None:
                process.kill()
            process.wait()


def _disable_influx_monitoring_in_process_list():
    process_list = _load_process_list()
    influx_process = _get_influx_process(process_list)
    _update_monitoring_environment_variable(influx_process)
    _write_process_list(process_list)


def _load_process_list():
    if not os.path.exists(PROCESS_LIST_JSON):
        raise ProcessListDoesNotExist()

    with open(PROCESS_LIST_JSON, 'r') as process_list_file:
        try:
            process_list = json.load(process_list_file, object_pairs_hook=OrderedDict)
        except Exception:
            raise ProcessListHasInvalidJson()
    
    return process_list


def _write_process_list(process_list):
    try:
        with open(PROCESS_LIST_JSON, 'w') as process_list_file:
            json.dump(process_list, process_list_file, indent=4)
    except Exception as error:
        raise ProcessListRewriteFailed()


def _get_influx_process(process_list):
    try:
        processes = process_list['application']['processes']
    except KeyError:
        raise ProcessListMalformed(['application', 'processes'])

    influx_processes = [
        process for process in processes
        if process['label'] == 'InfluxDB'
    ]

    if len(influx_processes) != 1:
        raise ProcessListInvalidInfluxEntriesCount(len(influx_processes))

    return influx_processes[0]


def _update_monitoring_environment_variable(influx_process):
    environment = influx_process.setdefault('environment', [])

    try:
        monitor_index = _get_monitoring_environment_index(environment)
    except ValueError:
        environment.append(MONITORING_ENV_VALUE)
    else:
        if environment[monitor_index] == MONITORING_ENV_VALUE:
            raise ProcessListMonitoringAlreadyDisabled()
        
        environment[monitor_index] = MONITORING_ENV_VALUE


def _get_monitoring_environment_index(environment):
    def _is_monitoring_variable(variable):
        return variable.startswith(MONITORING_ENV_VAR)

    return _index_of_first(environment, _is_monitoring_variable)


def _disable_monitoring_in_influx_config():
    config = _load_config_with_globals(INFLUX_CONFIG)

    with _suppress(ConfigParser.NoSectionError, ConfigParser.NoOptionError):
        if config.get('monitor', 'store-enabled') == 'false':
            raise InfluxConfigMonitoringAlreadyDisabled()

    with _suppress(ConfigParser.DuplicateSectionError):
        config.add_section('monitor')

    config.set('monitor', 'store-enabled', 'false')

    _write_config_with_globals(INFLUX_CONFIG, config)


def _load_config_with_globals(path, global_name='__global__'):
    if not os.path.exists(INFLUX_CONFIG):
        if not os.path.exists(INFLUXDB_DIR) or len(os.listdir(INFLUXDB_DIR)) == 0:
            raise NoSignOfInfluxHavingRun()
        else:
            raise InfluxDataExistsWithNoConfig()

    config = ConfigParser.ConfigParser(allow_no_value=True)

    with open(path) as raw_lines:
        lines_with_header = itertools.chain(
            ['[{}]\n'.format(global_name)],
            raw_lines
        )


        class FakeFile:
            def __init__(self, lines):
                self._lines = lines

            def readline(self):

                try:
                    line = next(self._lines)
                    if len(line.strip()) == 0:
                        return line
                    else:
                        return line.strip()
                except StopIteration:
                    return ''

        try:
            config.readfp(FakeFile(lines_with_header)) 
        except Exception:
            raise InfluxConfigHasInvalidIni()

    return config


def _write_config_with_globals(path, config, global_name='__global__'):
    try:
        with open(INFLUX_CONFIG, 'w') as out_file:
            for key, value in config.items(global_name):
                out_file.write('{} = {}\n'.format(key, value))
            
            if len(config.items(global_name)) > 0:
                out_file.write('\n')

            config.remove_section(global_name)
            config.write(out_file)
    except Exception:
        raise InfluxConfigRewriteFailed()
    

def _index_of_first(lst, pred):
    for i,v in enumerate(lst):
        if pred(v):
            return i
    raise ValueError


@contextlib.contextmanager
def _suppress(*exception_types):
    try:
        yield
    except exception_types:
        pass


@contextlib.contextmanager
def _devnull():
    with open(os.devnull, 'w') as DEVNULL:
        yield DEVNULL

